from typing import Any, List, Optional

COMMON_P: str
WORDS: Any
COMMON_WORDS: Any

def sentence() -> str: ...
def paragraph() -> str: ...
def paragraphs(count: int, common: bool = ...) -> List[str]: ...
def words(count: int, common: bool = ...) -> str: ...
