from typing import Any, Callable, List, Optional, Tuple

def get_func_args(func: Callable) -> List[str]: ...
def get_func_full_args(func: Callable) -> List[Tuple[str]]: ...
def func_accepts_kwargs(func: Callable) -> bool: ...
def func_accepts_var_args(func: Callable) -> bool: ...
def func_has_no_args(func: Any): ...
def func_supports_parameter(func: Callable, parameter: str) -> bool: ...
